import asyncio
import logging
import uuid
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from contextlib import contextmanager
from typing import Callable, Optional, Union

from .helpers import is_test

process_pool = ProcessPoolExecutor()
thread_pool = ThreadPoolExecutor(thread_name_prefix='run.py thread_pool')
running_processes = []  # NOTE is used in rosys.test.Runtime to advance time slower until computation is done
log = logging.getLogger('rosys.run')


async def io_bound(callback: Callable, *args: any):
    loop = asyncio.get_running_loop()
    try:
        return await loop.run_in_executor(thread_pool, callback, *args)
    except RuntimeError as e:
        if 'cannot schedule new futures after shutdown' not in str(e):
            raise
    except asyncio.exceptions.CancelledError:
        pass


async def cpu_bound(callback: Callable, *args: any):
    with cpu():
        try:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(process_pool, callback, *args)
        except RuntimeError as e:
            if 'cannot schedule new futures after shutdown' not in str(e):
                raise
        except asyncio.exceptions.CancelledError:
            pass


@contextmanager
def cpu():
    id = str(uuid.uuid4())
    running_processes.append(id)
    try:
        yield
    finally:
        running_processes.remove(id)


async def sh(command: Union[list[str], str], timeout: Optional[float] = 1) -> str:
    '''executes a shell command
    command: a sequence of program arguments as subprocess.Popen requires
    returns: stdout
    '''
    cmd_str = command if isinstance(command, str) else ' '.join(command)
    #log.info(f'running sh command "{cmd_str}"')
    try:
        proc = await asyncio.create_subprocess_shell(
            f'{cmd_str}' if timeout is None else f'timeout {timeout} {cmd_str}',
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )
    except:
        log.exception(f'"{cmd_str}" failed')
        return 'could not execute "{cmd_str}"'
    else:
        try:
            stdout, *_ = await proc.communicate()
        except asyncio.exceptions.CancelledError:
            return
        except:
            log.exception(f'"{cmd_str}" failed; waiting for process to finish')
            proc.kill()
            return 'could not execute "{cmd_str}"'
    #log.info(f'done executing "{cmd_str}"')
    return stdout.decode()


def tear_down():
    log.info('teardown thread_pool')
    thread_pool.shutdown(wait=True, cancel_futures=True)
    if not is_test:
        log.info('teardown process_pool')
        process_pool.shutdown(wait=True, cancel_futures=True)
    log.info('teardown complete')



from enum import Enum
from typing import Optional

from pydantic import Field
from rosys.world.rectangle import Rectangle

from .camera import Camera
from .image import ImageSize


class ImageRotation(str, Enum):
    NONE: int = 0
    RIGHT: int = 90
    UPSIDE_DOWN: int = 180
    LEFT: int = 270


class UsbCamera(Camera):
    active: bool = True
    detect: bool = False
    color: Optional[str] = Field(
        None,
        description='a color code to identify the camera'
    )
    resolution: Optional[ImageSize] = Field(
        None,
        description='physical resolution of the camera which should be used;'
        'camera may go into error state with wrong values'
    )
    auto_exposure: Optional[bool] = Field(
        True,
        description='toggles auto exposure'
    )
    exposure: Optional[float] = Field(
        None,
        description='manual exposure of the camera (between 0-1); set auto_exposure to False for this value to take effect'
    )
    rotation: ImageRotation = Field(
        ImageRotation.NONE,
        description='rotation which should be applied after grabbing and cropping'
    )
    fps: Optional[int] = Field(
        None,
        description='current frames per second (read only)'
    )
    crop: Optional[Rectangle] = Field(
        None,
        description='region to crop on the original resolution before rotation'
    )

    @property
    def image_resolution(self) -> Optional[ImageSize]:
        if self.resolution is None:
            return None
        width = self.crop.width if self.crop else self.resolution.width
        height = self.crop.height if self.crop else self.resolution.height
        if self.rotation == ImageRotation.LEFT or self.rotation == ImageRotation.RIGHT:
            return ImageSize(width=height, height=width)
        else:
            return self.resolution.copy()



from .planner_process import (PlannerCommand, PlannerGrowMapCommand, PlannerObstacleDistanceCommand, PlannerProcess,
                              PlannerResponse, PlannerSearchCommand, PlannerTestCommand)



import io
import random
import time
from typing import Optional

import PIL as pil
import rosys
from rosys.world.image import ImageSize

from ..world import Image, UsbCamera
from .actor import Actor


class UsbCameraSimulator(Actor):
    interval: float = 1

    async def step(self):
        await super().step()

        for camera in self.world.usb_cameras.values():
            if not camera.active:
                continue
            assert camera.image_resolution is not None, 'simulated USB cameras should have an image resolution'
            image = Image(time=self.world.time, camera_id=camera.id, size=camera.image_resolution)
            if rosys.is_test:
                image.data = b'test data'
            else:
                image.data = await rosys.run.cpu_bound(self.create_image_data, camera)
            camera.images.append(image)

    @staticmethod
    def create(uid: str, width: int = 800, height: int = 600, color: Optional[str] = None) -> UsbCamera:
        color = color or f'#{random.randint(0, 0xffffff):06x}'
        return UsbCamera(id=uid, resolution=ImageSize(width=width, height=height), connected=True, color=color)

    @staticmethod
    def create_calibrated(uid: str, width: int = 800, height: int = 600, color: Optional[str] = None,
                          x: float = 0, y: float = 0, z: float = 1,
                          yaw: float = 0, tilt_x: float = 0, tilt_y: float = 0) -> UsbCamera:
        camera = UsbCameraSimulator.create(uid, width, height, color)
        camera.set_perfect_calibration(x, y, z, yaw, tilt_x, tilt_y, width, height)
        return camera

    @staticmethod
    def create_image_data(camera: UsbCamera):
        size = camera.image_resolution
        img = pil.Image.new('RGB', size=(size.width, size.height), color=camera.color)
        d = pil.ImageDraw.Draw(img)
        text = f'{camera.id}: {time.time()}'
        d.text((img.width / 2 - len(text) * 3 + 0, img.height / 2 - 5), text, fill=(0, 0, 0))
        d.text((img.width / 2 - len(text) * 3 + 1, img.height / 2 - 4), text, fill=(255, 255, 255))
        img_byte_arr = io.BytesIO()
        img.save(img_byte_arr, format='JPEG')
        return img_byte_arr.getvalue()



from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import Callable

import rosys


@dataclass
class AppButton:
    icon: str
    state: str = 'enabled'
    visible: bool = True
    pressed: Callable[[], None] = lambda: None
    released: Callable[[], None] = lambda: None
    canceled: Callable[[], None] = lambda: None

    def get_properties(self) -> list[str]:
        return [
            f'/icon {self.icon}',
            f'/state {self.state}',
            f'/visible {"true" if self.visible else "false"}',
        ]

    def __str__(self) -> str:
        return f'AppButton({self.icon},{self.state},{self.visible})'


class AppControls():

    def __init__(self, robot_brain: rosys.hardware.RobotBrain) -> None:
        self.robot_brain = robot_brain
        self.main_buttons: dict[str, AppButton] = {}
        self.extra_buttons: dict[str, AppButton] = {}

    async def set_info(self, msg: str) -> None:
        '''replace constantly shown info text on mobile defice'''
        await self.robot_brain.send(f'bluetooth.send("PUT /info {msg}")')

    async def notify(self, msg: str) -> None:
        '''show notification as Snackbar message on mobile device'''
        await self.robot_brain.send(f'bluetooth.send("POST /notification {msg}")')

    async def parse(self, line: str) -> None:
        if line.startswith('"'):
            line = line[1:-1]
        if line.startswith('app: '):
            line = line[5:]
            if line == 'connected':
                await self.sync()
                rosys.event.emit(rosys.event.Id.APP_CONNECTED)
            elif line.startswith('PUT /button/') and '/action' in line:
                cmd = line.split(' ')
                path = cmd[1].split('/')
                name = path[-2]
                group = path[-3]
                action = cmd[2]
                buttons = self.main_buttons if group == 'main' else self.extra_buttons
                if action == 'pressed':
                    await self._invoke(buttons[name].pressed)
                if action == 'released':
                    await self._invoke(buttons[name].released)
                if action == 'canceled':
                    await self._invoke(buttons[name].canceled)

    async def sync(self):
        await self.send('PUT', lambda b: b.get_properties())

    async def clear(self):
        await self.send('DELETE')
        self.main_buttons.clear()

    async def send(self, method: str, get_properties: Callable[[AppButton], list[str]] = lambda b: ['']):
        async def run(group: str, buttons: list[AppButton]) -> None:
            for name, button in buttons.items():
                for prop in get_properties(button):
                    cmd = f'bluetooth.send("{method} /button/{group}/{name}{prop}")'
                    await self.robot_brain.send(cmd)
        await run('main', self.main_buttons)
        await run('exra', self.extra_buttons)

    async def _invoke(self, callback: Callable):
        if inspect.iscoroutinefunction(callback):
            await callback()
        else:
            callback()



from .app_controls import AppButton, AppControls
from .communicating_hardware import CommunicatingHardware
from .hardware import Hardware
from .robot_brain import RobotBrain
from .robot_brain_legacy import RobotBrainLegacy
from .simulated_hardware import SimulatedHardware



from ..world import Velocity, World
from .hardware import Hardware


class SimulatedHardware(Hardware):

    def __init__(self, world: World):
        super().__init__(world)
        self.linear_velocity: float = 0
        self.angular_velocity: float = 0

    async def configure(self):
        await super().configure()
        raise Exception('there is no communication for configuring hardware')

    async def restart(self):
        await super().restart()
        raise Exception('there is no communication for restarting hardware')

    async def drive(self, linear: float, angular: float):
        await super().drive(linear, angular)
        self.linear_velocity = linear
        self.angular_velocity = angular

    async def stop(self):
        await super().stop()
        self.linear_velocity = 0
        self.angular_velocity = 0

    async def update(self):
        await super().update()
        self.world.robot.odometry.append(Velocity(
            linear=self.linear_velocity,
            angular=self.angular_velocity,
            time=self.world.time,
        ))



from __future__ import annotations

import io
import os
from datetime import datetime
from glob import glob
from pathlib import Path

import humanize
import rosys
from PIL import Image, ImageDraw, ImageFont
from rosys.actors import Actor

from ..world import World

rosys_dir = os.path.dirname(os.path.dirname(__file__))
image_font = ImageFont.truetype(f'{rosys_dir}/RobotoMono-Medium.ttf', 12)


class TimelapsRecorder(Actor):
    interval: float = 1
    world: World
    storage_path: str = os.path.expanduser('~/.rosys/timelapse')

    def __init__(self) -> None:
        super().__init__()
        os.makedirs(self.storage_path, exist_ok=True)
        self.cover_font = ImageFont.truetype(f'{rosys_dir}/RobotoMono-Medium.ttf', 100)

    async def save(self, image: rosys.Image) -> None:
        await rosys.run.cpu_bound(save_image, image, self.storage_path)

    async def compress_video(self) -> None:
        jpgs = sorted(glob(f'{self.storage_path}/*.jpg'))
        if len(jpgs) < 2:
            return
        start = datetime.fromtimestamp(int(Path(jpgs[0]).stem))
        end = datetime.fromtimestamp(int(Path(jpgs[-1]).stem))
        self.log.info(f'creating video from {start} to {end}')
        duration = humanize.naturaldelta(end-start)
        for i in range(20):
            self.create_cover(start.strftime("%d.%m.%Y %H:%M:%S"), duration, f'{self.storage_path}/0_cover_{i}.jpg')
        id = start.strftime('%Y%m%d_%H-%M-%S_' + duration.replace(' ', '_'))
        target_dir = self.storage_path + '/' + id
        os.mkdir(target_dir)
        await rosys.run.sh(['mv', self.storage_path + '/*.jpg', target_dir])
        cmd = f'nice -n 19 ffmpeg -hide_banner -r 10 -pattern_type glob -i "{target_dir}/*.jpg" -s 1600x1200 -vcodec libx264 -crf 18 -preset slow -pix_fmt yuv420p -y {target_dir}/{id}.mp4; mv {target_dir}/*mp4 {self.storage_path}; rm -r {target_dir};'
        rosys.task_logger.create_task(rosys.run.sh(cmd, timeout=None))

    def clear_jpegs(self):
        files = glob(f'{self.storage_path}/**/*.jpg', recursive=True)
        for f in files:
            os.remove(f)

    def create_cover(self, title: str, subtilte: str, dest: str):
        img = Image.new('RGB', (1600, 1200), 'black')
        draw = ImageDraw.Draw(img)
        w, h = draw.textsize(title, font=self.cover_font)
        draw.text((img.width / 2 - w / 2, img.height / 2 - h / 2 - 200), title, font=self.cover_font, fill='white')
        w, h = draw.textsize(subtilte, font=self.cover_font)
        draw.text((img.width / 2 - w / 2, img.height / 2 - h / 2 + 100), subtilte, font=self.cover_font, fill='white')
        img.save(dest)

# static to run cpu bound


def save_image(image: rosys.world.Image, path: str):
    img = Image.open(io.BytesIO(image.data))
    draw = ImageDraw.Draw(img)
    x = img.width - 300
    y = img.height - 18
    text = f'{datetime.fromtimestamp(image.time).strftime("%Y-%m-%d %H:%M:%S")}, cam {image.camera_id}'
    # shadow
    draw.text((x - 1, y - 1), text, font=image_font, fill=(0, 0, 0))
    draw.text((x + 1, y - 1), text, font=image_font, fill=(0, 0, 0))
    draw.text((x - 1, y + 1), text, font=image_font, fill=(0, 0, 0))
    draw.text((x + 1, y + 1), text, font=image_font, fill=(0, 0, 0))
    draw.text((x, y), text, font=image_font, fill=(255, 255, 255))
    dest = path + f'/{int(image.time)}.jpg'
    img.save(dest, "JPEG")



from functools import lru_cache
from typing import Optional

import networkx as nx
import numpy as np
from scipy import ndimage, spatial

from ...helpers import angle
from ...world import Area, Obstacle, PathSegment, Point, Pose, Spline
from .delaunay_pose_group import DelaunayPoseGroup
from .fast_spline import FastSpline
from .grid import Grid
from .obstacle_map import ObstacleMap


class DelaunayPlanner:
    def __init__(self, robot_outline: list[tuple[float, float]]) -> None:
        self.robot_outline = robot_outline
        self.areas: list[Area] = []
        self.obstacles: list[Obstacle] = []
        self.obstacle_map: Optional[ObstacleMap] = None
        self.tri_points: Optional[np.array] = None
        self.tri_mesh: Optional[spatial.Delaunay] = None
        self.pose_groups: Optional[list[DelaunayPoseGroup]] = None
        self.graph: Optional[nx.DiGraph] = None

    def update_map(self, areas: list[Area], obstacles: list[Obstacle], additional_points: list[Point],
                   deadline: float) -> None:
        if self.obstacle_map and \
                self.areas == areas and \
                self.obstacles == obstacles and \
                all(self.obstacle_map.grid.contains(point, padding=1.0) for point in additional_points):
            return
        self.areas = areas
        self.obstacles = obstacles
        self._create_obstacle_map(additional_points, deadline)
        self._create_graph()

    def grow_map(self, points: list[Point], deadline: float) -> None:
        if self.obstacle_map is not None and \
                all(self.obstacle_map.grid.contains(point, padding=1.0) for point in points):
            return
        if self.obstacle_map is not None:
            bbox = self.obstacle_map.grid.bbox
            points.append(Point(x=bbox[0],         y=bbox[1]))
            points.append(Point(x=bbox[0]+bbox[2], y=bbox[1]))
            points.append(Point(x=bbox[0],         y=bbox[1]+bbox[3]))
            points.append(Point(x=bbox[0]+bbox[2], y=bbox[1]+bbox[3]))
        self._create_obstacle_map(points, deadline)
        self._create_graph()

    def _create_obstacle_map(self, additional_points: list[Point], deadline: float) -> None:
        points = [p for obstacle in self.obstacles for p in obstacle.outline]
        points += [p for area in self.areas for p in area.outline]
        points += additional_points
        grid = Grid.from_points(points, pixel_size=0.1, num_layers=36, padding=1.0)
        self.obstacle_map = ObstacleMap.from_world(self.robot_outline, self.areas, self.obstacles, grid, deadline)

    def _create_graph(self):
        min_x, min_y, size_x, size_y = self.obstacle_map.grid.bbox
        resolution = 1.0
        X, Y = np.meshgrid(np.arange(min_x, min_x + size_x - resolution / 2, resolution),
                           np.arange(min_y, min_y + size_y, resolution * np.sqrt(3) / 2))
        X[::2] += resolution / 2
        rows, cols = self.obstacle_map.grid.to_grid(X.flatten(), Y.flatten())
        keep = np.reshape([not all(self.obstacle_map.stack[int(np.round(row)), int(np.round(col)), :])
                           for row, col in zip(rows, cols)], X.shape)
        distance = ndimage.distance_transform_edt(1 - self.obstacle_map.map) * self.obstacle_map.grid.pixel_size
        D = ndimage.map_coordinates(distance, [[rows], [cols]], order=0).reshape(X.shape)
        keep[1::2, :] = np.logical_and(keep[1::2, :], D[1::2, :] < 2)
        keep[::4, 1::2] = np.logical_and(keep[::4, 1::2], D[::4, 1::2] < 2)
        keep[2::4, ::2] = np.logical_and(keep[2::4, ::2], D[2::4, ::2] < 2)
        self.tri_points = np.stack((X[keep], Y[keep]), axis=1)

        self.tri_mesh = spatial.Delaunay(self.tri_points)
        self.pose_groups = [
            DelaunayPoseGroup(
                index=i,
                point=Point(x=self.tri_points[i, 0], y=self.tri_points[i, 1]),
                neighbor_indices=_tri_neighbors(self.tri_mesh, i),
                poses=[
                    Pose(x=point[0], y=point[1], yaw=np.arctan2(neighbor[1] - point[1], neighbor[0] - point[0]))
                    for neighbor in self.tri_points[_tri_neighbors(self.tri_mesh, i)]
                ]
            )
            for i, point in enumerate(self.tri_points)
        ]

        self.graph = nx.DiGraph()
        for g, group in enumerate(self.pose_groups):
            for p in range(len(group.poses)):
                self.graph.add_node((g, p))
        for g, group in enumerate(self.pose_groups):
            for p, (pose, g_) in enumerate(zip(group.poses, group.neighbor_indices)):
                for p_, pose_ in enumerate(self.pose_groups[g_].poses):
                    if abs(angle(pose.yaw, pose_.yaw + np.pi)) < 0.01:
                        continue  # NOTE: avoid 180-degree turns
                    x, y, yaw = _generate_poses(self.obstacle_map.grid, pose, pose_)
                    if not self.obstacle_map.test(x, y, yaw).any():
                        length = np.sum(np.sqrt(np.diff(x)**2 + np.diff(y)**2))
                        self.graph.add_edge((g, p), (g_, p_), backward=False, weight=length)
                        if ((g_, p_), (g, p)) not in self.graph.edges:
                            self.graph.add_edge((g_, p_), (g, p), backward=True, weight=1.2*length)

    def search(self, start: Pose, goal: Pose) -> list[PathSegment]:
        first_segment, g, p = _find_terminal_segment(self.obstacle_map, self.pose_groups, start, True)
        last_segment, g_, p_ = _find_terminal_segment(self.obstacle_map, self.pose_groups, goal, False)

        path: list[PathSegment] = []
        path.append(first_segment)
        last_g, last_p = g, p
        try:
            for next_g, next_p in nx.shortest_path(self.graph, (g, p), (g_, p_), weight='weight')[1:]:
                last_pose = self.pose_groups[last_g].poses[last_p]
                next_pose = self.pose_groups[next_g].poses[next_p]
                backward = self.graph.edges[((last_g, last_p), (next_g, next_p))]['backward']
                spline = Spline.from_poses(last_pose, next_pose, backward=backward)
                path.append(PathSegment(spline=spline, backward=backward))
                last_g, last_p = next_g, next_p
        except nx.exception.NetworkXNoPath:
            pass
        path.append(last_segment)

        while True:
            for s in range(len(path) - 1):
                new_start = Pose(
                    x=path[s].spline.start.x,
                    y=path[s].spline.start.y,
                    yaw=path[s].spline.yaw(0) + (np.pi if path[s].backward else 0),
                )
                new_end = Pose(
                    x=path[s+1].spline.end.x,
                    y=path[s+1].spline.end.y,
                    yaw=path[s+1].spline.yaw(1) + (np.pi if path[s+1].backward else 0),
                )
                if abs(angle(new_start.yaw, new_end.yaw + np.pi)) < 0.01:
                    continue
                shortcuts = []
                for new_backward in [False, True]:
                    new_spline = Spline.from_poses(new_start, new_end, backward=new_backward)
                    if not _is_healthy(new_spline):
                        continue
                    if self.obstacle_map.test_spline(new_spline, new_backward):
                        continue
                    combined_length = _estimate_length(path[s].spline) + _estimate_length(path[s + 1].spline)
                    if .9 * _estimate_length(new_spline) > combined_length:
                        continue
                    shortcuts.append(PathSegment(spline=new_spline, backward=new_backward))
                if any(shortcuts):
                    lengths = [_estimate_length(segment.spline) for segment in shortcuts]
                    path[s] = shortcuts[np.argmin(lengths)]
                    del path[s+1]
                    break  # restart while loop
            else:
                break  # exit while loop

        return path


def _tri_neighbors(tri_mesh: spatial.Delaunay, vertex_index: int):
    return tri_mesh.vertex_neighbor_vertices[1][tri_mesh.vertex_neighbor_vertices[0][vertex_index]:
                                                tri_mesh.vertex_neighbor_vertices[0][vertex_index + 1]]


@lru_cache(maxsize=1000)
def _t_array(n: int) -> np.array:
    return np.linspace(0, 1, n)


@lru_cache(maxsize=10000)
def _generate_pose_offsets(grid: Grid, dx: float, dy: float, yaw: float, yaw_: float) -> tuple:
    spline = FastSpline(0, 0, yaw, dx, dy, yaw_, False)
    row0, col0, layer0 = grid.to_grid(0, 0, yaw)
    row1, col1, layer1 = grid.to_grid(dx, dy, yaw_)
    num_rows = int(abs(row1 - row0))
    num_cols = int(abs(col1 - col0))
    num_layers = int(abs(layer1 - layer0))
    t = _t_array(max(num_rows, num_cols, num_layers))
    return (spline.x(t), spline.y(t), spline.yaw(t) + [0, np.pi][spline.backward])


def _generate_poses(grid: Grid, pose: Pose, pose_: Pose) -> tuple:
    dx, dy, yaw = _generate_pose_offsets(grid, pose_.x - pose.x, pose_.y - pose.y, pose.yaw, pose_.yaw)
    return pose.x + dx, pose.y + dy, yaw


def _estimate_length(spline: Spline) -> float:
    dx = np.diff([spline.x(t) for t in np.linspace(0, 1, 10)])
    dy = np.diff([spline.y(t) for t in np.linspace(0, 1, 10)])
    return np.sum(np.sqrt(dx**2 + dy**2))


def _is_healthy(spline: Spline, curvature_limit: float = 10.0) -> bool:
    return np.abs(spline.max_curvature()) < curvature_limit


def _find_terminal_segment(obstacle_map: ObstacleMap, pose_groups: list[DelaunayPoseGroup],
                           terminal_pose: Pose, first: bool) -> tuple[PathSegment, int, int]:
    terminal_point = terminal_pose
    group_distances = [g.point.distance(terminal_point) for g in pose_groups]
    group_indices = np.argsort(group_distances)
    for g, group in zip(group_indices, np.array(pose_groups)[group_indices]):
        best_result = None
        best_length = np.inf
        for p, pose in enumerate(group.poses):
            for backward in [False, True]:
                poses = (terminal_pose, pose) if first else (pose, terminal_pose)
                spline = Spline.from_poses(*poses, backward=backward)
                if _is_healthy(spline) and not obstacle_map.test_spline(spline, backward):
                    length = _estimate_length(spline)
                    if length < best_length:
                        best_length = length
                        best_result = (PathSegment(spline=spline, backward=backward), g, p)
        if best_result is not None:
            return best_result
    raise RuntimeError('could not find terminal segment')



import datetime
import time

import humanize
from pydantic import BaseModel, PrivateAttr

from .. import event
from ..helpers import is_test
from .area import Area
from .camera import Camera
from .obstacle import Obstacle
from .robot import Robot
from .upload import Upload
from .usb_camera import UsbCamera


class World(BaseModel):
    robot: Robot = Robot()
    _time: float = PrivateAttr(default_factory=time.time)
    obstacles: dict[str, Obstacle] = {}
    areas: dict[str, Area] = {}
    notifications: list[tuple[float, str]] = []
    usb_cameras: dict[str, UsbCamera] = {}
    upload: Upload = Upload()
    needs_backup: bool = False
    start_time: float = 0

    @property
    def time(self) -> float:
        return self._time if is_test else time.time()

    def set_time(self, value):
        assert is_test
        self._time = value

    @property
    def cameras(self) -> dict[str, Camera]:
        return self.usb_cameras

    @property
    def uptime(self) -> str:
        uptime = datetime.timedelta(seconds=self.time - self.start_time)
        return humanize.precisedelta(uptime)

    @property
    def lizard_offset(self) -> str:
        if self.robot.hardware_time is None:
            return '-'
        offset_ms = (self.time - self.robot.hardware_time) * 1000
        return f'{int(offset_ms):4} ms'

    async def add_usb_camera(self, camera: UsbCamera) -> None:
        self.usb_cameras[camera.id] = camera
        await event.call(event.Id.NEW_CAMERA, camera)



from dataclasses import dataclass
from typing import Optional
from uuid import uuid4

import numpy as np
import rosys
from rosys.actors.detector import Autoupload

from ..world import Detections, Image, World
from . import Detector


@dataclass
class SimulatedObject:
    category_name: str
    position: rosys.world.Point3d
    size: Optional[tuple[float]] = None
    uuid: str = str(uuid4())


class DetectorSimulator(Detector):
    world: World
    noisy_image_points: bool = True

    def __init__(self) -> None:
        super().__init__()
        self.blocked_cameras: set[str] = set()
        self.simulated_objects: list[SimulatedObject] = []
        self.uploaded: list[rosys.world.Image] = []

    @property
    def is_connected(self):
        return True

    async def detect(self, image: Image, autoupload: Autoupload = Autoupload.FILTERED) -> Optional[Detections]:
        is_blocked = image.camera_id in self.blocked_cameras
        await rosys.sleep(0.4)
        image.detections = Detections()
        if not is_blocked:
            self.update_simulated_objects(image)
            self.detect_from_simulated_objects(image)
        rosys.event.emit(rosys.event.Id.NEW_DETECTIONS, image)
        return image.detections

    def update_simulated_objects(self, image: Image) -> None:
        pass

    def detect_from_simulated_objects(self, image: Image):
        if image.camera_id not in self.world.cameras:
            return
        camera = self.world.cameras[image.camera_id]
        for object in self.simulated_objects:
            image_point = camera.calibration_simulation.project_to_image(object.position)
            if not (0 <= image_point.x < image.size.width and 0 <= image_point.y < image.size.height):
                continue

            if object.size is None:
                detection = rosys.world.PointDetection(
                    category_name=object.category_name,
                    model_name='simulation',
                    confidence=1.0,
                    x=image_point.x,
                    y=image_point.y,
                    uuid=object.uuid,
                )
                if self.noisy_image_points:
                    detection.x += np.random.randn()
                    detection.y += np.random.randn()
                image.detections.points.append(detection)

            else:
                world_points = np.array([
                    [object.position.x + dx, object.position.y + dy, object.position.z + dz]
                    for dx in [-object.size[0] / 2, object.size[0] / 2]
                    for dy in [-object.size[1] / 2, object.size[1] / 2]
                    for dz in [-object.size[2] / 2, object.size[2] / 2]
                ])
                image_points = camera.calibration_simulation.project_array_to_image(world_points)
                detection = rosys.world.BoxDetection(
                    category_name=object.category_name,
                    model_name='simulation',
                    confidence=1.0,
                    x=image_points[:, 0].min(),
                    y=image_points[:, 1].min(),
                    width=image_points[:, 0].max() - image_points[:, 0].min(),
                    height=image_points[:, 1].max() - image_points[:, 1].min(),
                    uuid=object.uuid,
                )
                if self.noisy_image_points:
                    detection.x += np.random.randn()
                    detection.y += np.random.randn()
                    detection.width += np.random.randn()
                    detection.height += np.random.randn()
                image.detections.boxes.append(detection)

    async def upload(self, image: rosys.world.Image):
        self.uploaded.append(image)



import logging
from typing import Optional

import cv2
import numpy as np
import rosys
from aiocache import cached
from nicegui.ui import Ui
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.routing import Route

log = logging.getLogger('rosys.routes')


def shrink(factor: int, array: np.ndarray) -> bytes:
    img = cv2.imdecode(array, cv2.IMREAD_COLOR)[::factor, ::factor]
    return cv2.imencode('.jpg', img, [int(cv2.IMWRITE_JPEG_QUALITY), 60])[1].tobytes()


def setup(ui: Ui, runtime: rosys.Runtime):
    ui.add_route(Route('/world', lambda request, **_: Response(content=runtime.world.json(), media_type='text/json')))
    ui.add_route(Route('/export', lambda request, **_: JSONResponse(content=runtime.persistence.dump())))

    @cached(ttl=30)
    async def try_get_jpeg(cam_id: str, timestamp: str, shrink_factor: int) -> Optional[bytes]:
        camera = runtime.world.cameras.get(cam_id)
        if camera is None:
            return
        for image in reversed(camera.images):
            if str(image.time) == timestamp:
                jpeg = image.data
                if shrink_factor != 1:
                    array = np.frombuffer(image.data, dtype=np.uint8)
                    jpeg = await rosys.run.cpu_bound(shrink, shrink_factor, array)
                return jpeg

    async def get_image(request: Request, **_):
        try:
            jpeg = await try_get_jpeg(request.path_params['id'],
                                      request.path_params['timestamp'],
                                      int(request.query_params.get('shrink', 1)))
            if jpeg:
                return Response(content=jpeg, headers={'cache-control': 'max-age=7776000'}, media_type='image/jpeg')
            else:
                return Response(content='Not Found', status_code=404)
        except:
            log.exception('could not get image')
            raise

    ui.add_route(Route('/camera/{id}/{timestamp}', get_image))
    ui.add_route(Route('/camera/placeholder', lambda _: Response(
        content=rosys.world.Image.create_placeholder('no image').data,
        media_type='image/jpeg'
    )))



from pydantic import BaseModel

from .point import Point


class Detection(BaseModel):
    category_name: str
    model_name: str
    confidence: float
    x: float
    y: float
    uuid: str = ''  # NOTE this is used to keep track of simulated objects

    @property
    def cx(self) -> float:
        raise NotImplementedError()

    @property
    def cy(self) -> float:
        raise NotImplementedError()

    @property
    def center(self) -> Point:
        return Point(x=self.cx, y=self.cy)


class BoxDetection(Detection):
    width: float
    height: float

    @property
    def cx(self) -> float:
        return self.x + self.width / 2

    @property
    def cy(self) -> float:
        return self.y + self.height / 2

    def __str__(self) -> str:
        return f'x:{self.x:.0f} y: {self.y:.0f}, w: {self.width:.0f} h: {self.height:.0f} cat: {self.category_name} conf: {self.confidence:.2f}'

    def to_svg(self, shrink: int = 1) -> str:
        color = 'red'
        x = self.x / shrink
        y = self.y / shrink
        return f'<rect x="{x}" y="{y}" width="{self.width/shrink}" height="{self.height/shrink}" stroke-width="2" stroke="{color}" fill="none" />' \
            f'<text x="{x}" y="{y - 7}" text-anchor="start" stroke="{color}" fill="{color}" font-size="10">{self.category_name} ({int(self.confidence*100)}%)</text>'


class PointDetection(Detection):

    @property
    def cx(self) -> float:
        return self.x

    @property
    def cy(self) -> float:
        return self.y

    def __str__(self) -> str:
        return f'x:{self.x:.0f} y: {self.y:.0f} cat: {self.category_name} conf: {self.confidence:.2f}'

    def to_svg(self, shrink: int = 1) -> str:
        color = 'red'
        x = self.x / shrink
        y = self.y / shrink
        return f'<circle cx="{x}" cy="{y}" r="4" stroke-width="2" stroke="{color}" fill="none" />' \
            f'<text x="{x + 10}" y="{y + 4}" text-anchor="start" stroke="{color}" fill="{color}" font-size="12" font-weight="light">{self.category_name} ({int(self.confidence*100)}%)</text>'


class Detections(BaseModel):
    boxes: list[BoxDetection] = []
    points: list[PointDetection] = []

    def to_svg(self, shrink: int = 1) -> str:
        return '\n'.join(b.to_svg(shrink) for b in self.boxes) + '\n' + '\n'.join(p.to_svg(shrink) for p in self.points)



from datetime import datetime, timedelta
from typing import Coroutine, Optional

import socketio
import socketio.exceptions
from aenum import Enum, auto

from .. import event, task_logger
from ..helpers import sleep
from ..world import BoxDetection, Detections, Image, PointDetection
from .actor import Actor


class Autoupload(Enum, init='value __doc__'):
    '''Configures the auto-submitting of images to the Learning Loop'''

    def _generate_next_value_(name: str, start, count, last_values):
        '''uses enum name as value when calling auto()'''
        return name.lower()

    FILTERED = auto(), 'only submit images with novel detections and in an uncertainty range (this is the default)'
    DISABLED = auto(), 'no auto-submitting'
    ALL = auto(), 'submit all images which are run through the detector'


class Detector(Actor):
    interval: float = 1.0

    def __init__(self, port: int = 8004, name: str = 'detector'):
        super().__init__()
        self.sio = socketio.AsyncClient()
        self.is_detecting: bool = False
        self.next_image: Optional[Image] = None
        self.port = port
        self.name = name
        self.timeout_count = 0

        @self.sio.on('disconnect')
        def on_sio_disconnect():
            self.log.warning(f'sio disconnect on port {port}')

        @self.sio.on('connect_error')
        def on_sio_connect_error(err):
            self.log.warning(f'sio connect error on {port}: {err.message}')

    @property
    def is_connected(self):
        return self.sio.connected

    async def step(self):
        if not self.is_connected:
            self.log.info(f'trying reconnect {self.port}')
            if not await self.connect():
                self.log.exception(f'connection to {self.port} failed; trying again')
                await sleep(3.0)
                return

        await self.try_start_one_upload()
        await self.upload_priority_queue()

    async def connect(self) -> bool:
        try:
            url = f'ws://localhost:{self.port}'
            self.log.info(f'connecting to detector at {url}')
            await self.sio.connect(url, socketio_path='/ws/socket.io', wait_timeout=3.0)
            self.log.info('connected successfully')
            return True
        except:
            return False

    async def disconnect(self) -> Coroutine:
        await self.sio.disconnect()

    async def try_start_one_upload(self) -> Coroutine:
        if datetime.now() < self.world.upload.last_upload + timedelta(minutes=self.world.upload.minimal_minutes_between_uploads):
            return

        upload_images = self.world.upload.get_queued(self.name)
        if upload_images:
            task_logger.create_task(self.upload(upload_images[0]), name='upload_image')
            self.world.upload.queue.clear()  # old images should not be uploaded later when the robot is inactive
            self.world.upload.last_upload = datetime.now()

    async def upload_priority_queue(self) -> Coroutine:
        upload_images = self.world.upload.get_priority_queued(self.name)
        if upload_images:
            async def upload_priority_images():
                for image in upload_images:
                    await self.upload(image)
            task_logger.create_task(upload_priority_images(), name='upload_priority_images')
            self.world.upload.priority_queue.clear()

    async def upload(self, image: Image) -> Coroutine:
        try:
            self.log.info(f'uploading to {self.name}')
            await self.sio.emit('upload', {'image': image.data, 'mac': image.camera_id})
        except:
            self.log.exception(f'could not upload {image.id}')

    async def detect(self, image: Image, autoupload: Autoupload = Autoupload.FILTERED) -> Coroutine:
        '''Runs detections on the image. Afterwards the `image.detections` property is filled.'''
        if not self.is_connected:
            return

        self.next_image = image
        if self.is_detecting:
            return

        while self.next_image is not None:
            try:
                image = self.next_image
                self.next_image = None
                self.is_detecting = True
                result = await self.sio.call('detect', {
                    'image': image.data,
                    'mac': image.camera_id,
                    'autoupload': autoupload.value
                }, timeout=3)
                box_detections = [BoxDetection.parse_obj(d) for d in result.get('box_detections', [])]
                point_detections = [PointDetection.parse_obj(d) for d in result.get('point_detections', [])]
                image.detections = Detections(boxes=box_detections, points=point_detections)
            except socketio.exceptions.TimeoutError:
                self.log.exception(f'detection for {image.id} on {self.port} took to long')
                self.timeout_count += 1
            except:
                self.log.exception(f'could not detect {image.id}')
            else:
                self.timeout_count = 0
                event.emit(event.Id.NEW_DETECTIONS, image)
            finally:
                self.is_detecting = False
                if self.timeout_count > 5:
                    await self.disconnect()

    def __str__(self) -> str:
        return f'{type(self).__name__} ({"connected" if self.is_connected else "disconnected"})'



#!/usr/bin/env python3
from __future__ import annotations

import time

import numpy as np
import pylab as pl
from nicegui import ui
from rosys.actors.pathplanning import plot_tools as pt
from rosys.actors.pathplanning.delaunay_planner import DelaunayPlanner
from rosys.actors.pathplanning.planner_process import PlannerSearchCommand
from rosys.actors.pathplanning.robot_renderer import RobotRenderer
from rosys.world import Area, Obstacle, Point, Pose

seed = np.random.randint(0, 1000)
seed = 0
np.random.seed(seed)

cmd = PlannerSearchCommand(
    areas=[Area(id='main', outline=[Point(x=-5, y=-5), Point(x=25, y=-5), Point(x=25, y=25), Point(x=-5, y=25)])],
    obstacles=[
        Obstacle(id='0', outline=[Point(x=5, y=-5), Point(x=8, y=-5), Point(x=8, y=15), Point(x=5, y=15)]),
        Obstacle(id='1', outline=[Point(x=12, y=5), Point(x=15, y=5), Point(x=15, y=25), Point(x=12, y=25)]),
    ],
    start=Pose(x=np.random.uniform(-4, 4), y=np.random.uniform(-4, 4), yaw=np.random.uniform(-np.pi, np.pi)),
    goal=Pose(x=np.random.uniform(16, 24), y=np.random.uniform(16, 24), yaw=np.random.uniform(-np.pi, np.pi)),
    deadline=np.inf,
)
robot_outline = [(-0.5, -0.5), (0.5, -0.5), (0.75, 0), (0.5, 0.5), (-0.5, 0.5)]
planner = DelaunayPlanner(robot_outline)


def run():
    t = time.time()
    planner.update_map(cmd.areas, cmd.obstacles, [cmd.start, cmd.goal], deadline=time.time()+10.0)
    dt0 = time.time() - t

    t = time.time()
    path = planner.search(cmd.start, cmd.goal)
    dt1 = time.time() - t

    with plot:
        pl.clf()
        pl.title(f'map: {dt0:.3f} s, path: {dt1:.3f} s ({seed=})')
        pt.show_obstacle_map(planner.obstacle_map)
        pl.gca().invert_yaxis()
        pl.autoscale(False)
        pl.triplot(planner.tri_points[:, 0], planner.tri_points[:, 1], planner.tri_mesh.simplices, lw=0.1)

        pt.plot_path(path, 'C1')
        robot_renderer = RobotRenderer(robot_outline)
        pt.plot_robot(robot_renderer, (cmd.start.x, cmd.start.y, cmd.start.yaw), 'C0', lw=2)
        pt.plot_robot(robot_renderer, (cmd.goal.x, cmd.goal.y, cmd.goal.yaw), 'C0', lw=2)
        for step in path:
            for t in [0, 1]:
                yaw = step.spline.yaw(t) + np.pi if step.backward else step.spline.yaw(t)
                pt.plot_robot(robot_renderer, (step.spline.x(t), step.spline.y(t), yaw), 'C2', lw=1)


with ui.row():
    plot = ui.plot(figsize=(8, 8))
    ui.button('Re-run', on_click=run).props('icon=replay outline')
run()

ui.run()
